/*
 * file:     tklib.c
 * author:   Doug du Boulay from x11lib.c by JC, aka Carolyn Wasikowski
 *           and thence from pxw.c by R. Olthof
 * date:     28/10/99
 * purpose:  To build Tcl/Tk interfaces for PLOTX, PREVUE, PIG, GIP, PREABS  
 *           and EXAMPP. This is a FORTRAN-callable interface, 
 *           that supports its own XtalCanvas widget type based on tkCanvas
 *           for direct rendering with X11. It also enables multiple windows 
 *           and other tcl/tk widget constructs. The XtalCanvas item
 *           supports segments and retained graphics (modelled after GKS).
 *
 ************************************************************************
 *      Copyright (C) 2000, The Xtal System, The University of          *
 *            Western Australia, xtal@crystal.uwa.edu.au.               *
 *                                                                      *
 * Distributed under the GNU General Public Licence as published by     *
 * the Free Software Foundation, in the hope that it will be useful     *
 * but WITHOUT ANY WARRANTY. See the GNU General Public License         *
 * for more details.                                                    *
 *                                                                      *
 ************************************************************************
 * 
 */

/* uncomment this and recompile for thick lines for screen capture shots*/
/* #define SCREENSHOT */

#ifndef _XTALCANVAS
#define _XTALCANVAS


/*
 * include files
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <tk.h>
/*
#if ((TCL_MAJOR_VERSION != 8) && (TCL_MINOR_VERSION != 2))
#error "You need the Tcl 8.2.x source release from Scriptics for this"
#end
 */
/*  the following is Microsoft C++/ Digital Fortran90 specific */
/* __WIN32__ is defined explicitly from tk.h ! */
#ifdef __WIN32__
#else
#define __stdcall
#endif


#ifdef NO_STDLIB_H
#   include "../compat/stdlib.h"
#else
#   include <stdlib.h>
#endif
/*
  #define _INCLUDE_HPUX_SOURCE
  #include <sys/time.h>
  #undef  _INCLUDE_HPUX_SOURCE
 */
#ifdef VMS
/* define _VMS_V6_SOURCE */
#endif





/*----------------------------------------------------------------------*
 * Constant, structure, variable, and function declarations.
 *----------------------------------------------------------------------*/

/*
 * constant definitions
 */
#define NUM_COLORS	(240)		/* number of defined colors, starting
					 *  with color #0 */
#define NUM_PATTERNS	(9)		/* number of defined patterns, starting
					 *  with pattern #1 */
#define CMD_BUF_SIZE	(1000)		/* size of input command buffer--must
					 *  match buffrN macro definition */
#define STR_BUF_SIZE	(72)		/* size of input string buffer--set to
					 *  the size of NCAR in PV31 */
#define MAX_COLOR_COMP	(65535)		/* maximum value of an X color 
					 *  component (red, green, or blue) */
#define FULL_CIRCLE	(23040)		/* 64 * 360 degrees */

#define DefaultDashIndex	(2)
#define DefaultFillStyle	(FillSolid)
#define	DefaultLineWidth	(0)
#define	DefaultLineStyle	(LineSolid)
#define	DefaultPatternIndex	(2)

#define IgnoreLineWidth		(9999)
#define IgnoreLineStyle		(-1)

#define PatternIndexFlag	(-1)

#define InvalidForeground	(-1)

/*
 * definition of icon bitmap and stipple bitmaps
 */

#define pattern1_width 4		/* "hollow" */
#define pattern1_height 4
static char pattern1_bits[] = {
   0x00, 0x00, 0x00, 0x00};

#define pattern2_width 4		/* solid */
#define pattern2_height 4
static char pattern2_bits[] = {
   0xff, 0xff, 0xff, 0xff};

#define pattern3_width 6		/* dotted pattern */
#define pattern3_height 10
static char pattern3_bits[] = {
   0x03, 0x03, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00};

#define pattern4_width 1		/* horizontal lines */
#define pattern4_height 4
static char pattern4_bits[] = {
   0x01, 0x00, 0x00, 0x00};

#define pattern5_width 4		/* vertical lines */
#define pattern5_height 1
static char pattern5_bits[] = {
   0x01};

#define pattern6_width 4		/* horizontal and vertical hatching */
#define pattern6_height 4
static char pattern6_bits[] = {
   0x0f, 0x01, 0x01, 0x01};

#define pattern7_width 6		/* SW to NE diagonal lines */
#define pattern7_height 6
static char pattern7_bits[] = {
   0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

#define pattern8_width 6		/* NW to SE diagonal lines */
#define pattern8_height 6
static char pattern8_bits[] = {
   0x01, 0x02, 0x04, 0x08, 0x10, 0x20};

#define pattern9_width 6		/* diagonal hatching */
#define pattern9_height 6
static char pattern9_bits[] = {
   0x01, 0x22, 0x14, 0x08, 0x14, 0x22};


/*
 * type and variable definitions
 */
typedef union  _xCommandRec	*xCommand;	/* forward declaration */
typedef struct _xSegmentRec	*xSegment;	/* forward declaration */

/* bloody big pointer 
  typedef long long int  XtlPtr ;
 */
typedef  long  XtlPtr ;

typedef void (__stdcall *FortAuxRefSub)() ;
typedef void (__stdcall *FortMouseEventCB)(float *,float *) ;
typedef void (__stdcall *FortKeyEventCB)(float *,float *) ;
typedef void (__stdcall *FortAuxWinConfig)(int *,int *) ;
typedef void (__stdcall *FortAuxWinScroll)(float *,float *) ;
typedef void (__stdcall *FTimerCb)() ;

typedef struct _timerRec {                     /* a timer record for animation*/
        int                   id;              /* timer id number  */
        Tcl_TimerToken        token;
        FTimerCb              timerCallback; /* pointer to callback func */
struct _timerRec              *next;           /* point to next timer in list*/
        } timerRec ;



/* ############################### XTAL Segment Definitions ################ */

typedef enum {
	xAnyCommandType,			/* any command */
	xLinesCommandType,			/* draw polyline */
	xFillCommandType,			/* fill polygon */
	xMarkCommandType,			/* draw markers */
	xTextCommandType			/* draw text in segment */
	} xCommandType;				/* a command type */

typedef struct {
	xCommandType		type;		/* xAnyCommandType */
	xCommand		next;		/* next cmd in list */
	} xAnyCommandRec, *xAnyCommand;		/* used to determine cmd type */

typedef struct {
	xCommandType		type;		/* xLinesCommandType */
	xCommand		next;		/* next cmd in list */
	int			fg;		/* foreground color */
	int			ls;		/* line style */
	int			lw;		/* line width */
	int			n;		/* number of points */
	float			*x;		/* x-coords */
	float			*y;		/* y-coords */
	int			up;		/* 2 for pen-down, 3 for up */
	} xLinesCommandRec, *xLinesCommand;	/* info needed for lines cmd */

typedef struct {
	xCommandType		type;		/* xFillCommandType */
	xCommand		next;		/* next cmd in list */
	int			fg;		/* foreground color */
	int			fs;		/* fill style */
	int			n;		/* number of points */
	float			*x;		/* x-coords */
	float			*y;		/* y-coords */
	} xFillCommandRec, *xFillCommand;	/* info needed for fill cmd */

typedef struct {
	xCommandType		type;		/* xMarkCommandType */
	xCommand		next;		/* next cmd in list */
	int			fg;		/* foreground color */
	float			size;		/* size of markers */
	int			n;		/* number of markers */
	float			*x;		/* x-coords */
	float			*y;		/* y-coords */
	} xMarkCommandRec, *xMarkCommand;	/* info needed for mark cmd */

typedef struct {
        xCommandType            type;           /* xTextCommandType */
        xCommand                next;           /* next cmd in list */
        int                     fg;             /* foreground color */
        float                   size;           /* character size   */
        float                   delx;           /* x dimension char shift */
        float                   dely;           /* y dimension char shift */
        float                   x;              /* x-coords */
        float                   y;              /* y-coords */
        char                    *string;        /* the actual text */
        } xTextCommandRec, *xTextCommand;       /* info needed for mark cmd */

typedef union _xCommandRec {
	xAnyCommandRec		any;
	xLinesCommandRec	lines;
	xFillCommandRec		fill;
	xMarkCommandRec		mark;
	xTextCommandRec		text;
	} xCommandRec;				/* info needed for any cmd */

typedef struct _xSegmentRec {
	xSegment		next;		/* next segment in list */
	int			id;		/* this segment's id number */
	Bool			detect;		/* is this seg detectable? */
	Bool			hidden;		/* should segment be displayed*/
	xCommand		cmdList;	/* list of this seg's cmds */
	xCommand		lastCmd;	/* points to last cmd in list */
	float			xmin;		/* segment's extrema */
	float			xmax;
	float			ymin;
	float			ymax;
} xSegmentRec;				/* a group of commands */


/*
 * The record below describes a canvas widget.  It is made available
 * to the item procedures so they can access certain shared fields such
 * as the overall displacement and scale factor for the canvas.
 */

typedef struct XtalCanvas {
    Tk_Window tkwin;		/* Window that embodies the canvas.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up.*/
    Display *display;		/* Display containing widget;  needed, among
				 * other things, to release resources after
				 * tkwin has already gone away. */
    Tcl_Interp *interp;		/* Interpreter associated with canvas. */
    Tcl_Command widgetCmd;	/* Token for canvas's widget command. */

    /*
     * Information used when displaying widget:
     */

    int borderWidth;		/* Width of 3-D border around window. */
    Tk_3DBorder bgBorder;	/* Used for canvas background. */
    int relief;			/* Indicates whether window as a whole is
				 * raised, sunken, or flat. */
    int highlightWidth;		/* Width in pixels of highlight to draw
				 * around widget when it has the focus.
				 * <= 0 means don't draw a highlight. */
    XColor *highlightBgColorPtr;
				/* Color for drawing traversal highlight
				 * area when highlight is off. */
    XColor *highlightColorPtr;	/* Color for drawing traversal highlight. */
    int inset;			/* Total width of all borders, including
				 * traversal highlight and 3-D border.
				 * Indicates how much interior stuff must
				 * be offset from outside edges to leave
				 * room for borders. */
    GC pixmapGC;		/* Used to copy bits from a pixmap to the
				 * screen and also to clear the pixmap. */
    int width, height;		/* Dimensions to request for canvas window,
				 * specified in pixels. */
    int redrawX1, redrawY1;	/* Upper left corner of area to redraw,
				 * in pixel coordinates.  Border pixels
				 * are included.  Only valid if
				 * REDRAW_PENDING flag is set. */
    int redrawX2, redrawY2;	/* Lower right corner of area to redraw,
				 * in integer canvas coordinates.  Border
				 * pixels will *not* be redrawn. */
    int confine;		/* Non-zero means constrain view to keep
				 * as much of canvas visible as possible. */


    /*
     * Transformation applied to canvas as a whole:  to compute screen
     * coordinates (X,Y) from canvas coordinates (x,y), do the following:
     *
     * X = x - xOrigin;
     * Y = y - yOrigin;
     */

    int xOrigin, yOrigin;	/* Canvas coordinates corresponding to
				 * upper-left corner of window, given in
				 * canvas pixel units. */
    int drawableXOrigin, drawableYOrigin;
				/* During redisplay, these fields give the
				 * canvas coordinates corresponding to
				 * the upper-left corner of the drawable
				 * where items are actually being drawn
				 * (typically a pixmap smaller than the
				 * whole window). */

    /*
     * Information used for event bindings associated with items.
     */

    Tk_BindingTable bindingTable;
				/* Table of all bindings currently defined
				 * for this canvas.  NULL means that no
				 * bindings exist, so the table hasn't been
				 * created.  Each "object" used for this
				 * table is either a Tk_Uid for a tag or
				 * the address of an item named by id. */

    /*
     * Information used for managing scrollbars:
     */

    char *xScrollCmd;		/* Command prefix for communicating with
				 * horizontal scrollbar.  NULL means no
				 * horizontal scrollbar.  Malloc'ed*/
    char *yScrollCmd;		/* Command prefix for communicating with
				 * vertical scrollbar.  NULL means no
				 * vertical scrollbar.  Malloc'ed*/
    int scrollX1, scrollY1, scrollX2, scrollY2;
				/* These four coordinates define the region
				 * that is the 100% area for scrolling (i.e.
				 * these numbers determine the size and
				 * location of the sliders on scrollbars).
				 * Units are pixels in canvas coords. */
    char *regionString;		/* The option string from which scrollX1
				 * etc. are derived.  Malloc'ed. */
    int xScrollIncrement;	/* If >0, defines a grid for horizontal
				 * scrolling.  This is the size of the "unit",
				 * and the left edge of the screen will always
				 * lie on an even unit boundary. */
    int yScrollIncrement;	/* If >0, defines a grid for horizontal
				 * scrolling.  This is the size of the "unit",
				 * and the left edge of the screen will always
				 * lie on an even unit boundary. */

    /*
     * Information used for scanning:
     */

    int scanX;			/* X-position at which scan started (e.g.
				 * button was pressed here). */
    int scanXOrigin;		/* Value of xOrigin field when scan started. */
    int scanY;			/* Y-position at which scan started (e.g.
				 * button was pressed here). */
    int scanYOrigin;		/* Value of yOrigin field when scan started. */


    /*
     * Miscellaneous information:
     */

    Tk_Cursor cursor;		/* Current cursor for window, or None. */
    char *takeFocus;		/* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL. */
    double pixelsPerMM;		/* Scale factor between MM and pixels;
				 * used when converting coordinates. */
    int flags;			/* Various flags;  see below for
				 * definitions. */

/* Xtal modifications */

    Tcl_HashTable       eventtable;     /* limited key, mouse  table */

    Pixmap              pixmap;         /*      */
    Drawable		canvas;		/* X ID of drawable window/pixmap */
    int			doubleBuffered;	/*  */
    int   	        depth;		/* depth of the graphic */
    FortAuxRefSub       auxrefcom;      /* auxilliary refresh commands*/
    FortAuxWinConfig    auxwinconf;     /* auxilliary refresh commands*/
    FortAuxWinScroll    auxwinxscroll;  /* extra commands to call for xscroll*/
    FortAuxWinScroll    auxwinyscroll;  /* extra commands to call for yscroll*/
    Bool		aspectFixed;	/* unity = 1 free = 0 */
    float 		aspectRatio,	/* the imposed aspect ratio*/
                    	xCentre,	/* centre of graphic */
			yCentre,	/* centre of graphic */
			xScale,		/* imposed horizontal scale */
			yScale;		/* imposed vertical scale fact*/
    int			xOffset;
    int			yOffset;
    struct	{
		float	w;		/* width resize factor */
		float	h;		/* width resize factor */
		} resizeFactor;	        /*often lesser of width&height*/
    float               *saspect;       /* shadow width pointer */
    float               *sxmag;         /* shadow width pointer */
    float               *symag;         /* shadow width pointer */
    float               *sxcent;        /* shadow width pointer */
    float               *sycent;        /* shadow width pointer */
    xSegment		segmentList;	/* list of this frame's segs */
    xSegment       	openSegment;	/* Currently opened segment*/
    xSegment       	lastSegment;	/* Currently opened segment*/
    unsigned long	fgCol;		/* window foreground Color */
    int			fillStyle;
    int			dashIndex;	
    int			lineStyle;      /* solid 0 onoff 1 double 2 */
    int			lineIndex;      /* 1 2 3 4 5 */
    unsigned int	lineWidth;
    int			fillColor;
    int			lineColor;
    int			stippleStyle;
    int			stippleIndex;
    int			markerColor;
    float		markerSize;
    float		xLast;
    float		yLast;
    XPoint		pen;
    Bool		penUp;
    struct  XtalCanvas	*next;		/* next Graphic in list */
    struct  XtalCanvas	*prev;		/* prev Graphic in list */
} XtalCanvas ; 

/*
 * Flag bits for canvases:
 *
 * REDRAW_PENDING -		1 means a DoWhenIdle handler has already
 *				been created to redraw some or all of the
 *				canvas.
 * REDRAW_BORDERS - 		1 means that the borders need to be redrawn
 *				during the next redisplay operation.
 * REPICK_NEEDED -		1 means DisplayCanvas should pick a new
 *				current item before redrawing the canvas.
 * GOT_FOCUS -			1 means the focus is currently in this
 *				widget, so should draw the insertion cursor
 *				and traversal highlight.
 * CURSOR_ON -			1 means the insertion cursor is in the "on"
 *				phase of its blink cycle.  0 means either
 *				we don't have the focus or the cursor is in
 *				the "off" phase of its cycle.
 * UPDATE_SCROLLBARS -		1 means the scrollbars should get updated
 *				as part of the next display operation.
 * LEFT_GRABBED_ITEM -		1 means that the mouse left the current
 *				item while a grab was in effect, so we
 *				didn't change canvasPtr->currentItemPtr.
 * REPICK_IN_PROGRESS -		1 means PickCurrentItem is currently
 *				executing.  If it should be called recursively,
 *				it should simply return immediately.
 */

#define REDRAW_PENDING		1
#define REDRAW_BORDERS		2
#define REPICK_NEEDED		4
#define GOT_FOCUS		8
#define CURSOR_ON		0x10
#define UPDATE_SCROLLBARS	0x20
#define LEFT_GRABBED_ITEM	0x40
#define REPICK_IN_PROGRESS	0x100



/*
 * A hash table is kept to map from the string names of event
 * modifiers to information about those modifiers.  The structure
 * for storing this information, and the hash table built at
 * initialization time, are defined below.
 */

typedef struct {
    char *name;			/* Name of modifier. */
    int mask;			/* Button/modifier mask value,							 * such as Button1Mask. */
    int flags;			/* Various flags;  see below for
				 * definitions. */
} ModInfo;

/*
 * Flags for ModInfo structures:
 *
 * DOUBLE -		Non-zero means duplicate this event,
 *			e.g. for double-clicks.
 * TRIPLE -		Non-zero means triplicate this event,
 *			e.g. for triple-clicks.
 */

#define DOUBLE		1
#define TRIPLE		2

/*
 * The following special modifier mask bits are defined, to indicate
 * logical modifiers such as Meta and Alt that may float among the
 * actual modifier bits.
 */

#define META_MASK	(AnyModifier<<1)
#define ALT_MASK	(AnyModifier<<2)

static ModInfo modArray[] = {
    {"Control",		ControlMask,	0},
    {"Shift",		ShiftMask,	0},
    {"Lock",		LockMask,	0},
    {"Meta",		META_MASK,	0},
    {"M",		META_MASK,	0},
    {"Alt",		ALT_MASK,	0},
    {"B1",		Button1Mask,	0},
    {"Button1",		Button1Mask,	0},
    {"B2",		Button2Mask,	0},
    {"Button2",		Button2Mask,	0},
    {"B3",		Button3Mask,	0},
    {"Button3",		Button3Mask,	0},
    {"B4",		Button4Mask,	0},
    {"Button4",		Button4Mask,	0},
    {"B5",		Button5Mask,	0},
    {"Button5",		Button5Mask,	0},
    {"Mod1",		Mod1Mask,	0},
    {"M1",		Mod1Mask,	0},
    {"Command",		Mod1Mask,	0},
    {"Mod2",		Mod2Mask,	0},
    {"M2",		Mod2Mask,	0},
    {"Option",		Mod2Mask,	0},
    {"Mod3",		Mod3Mask,	0},
    {"M3",		Mod3Mask,	0},
    {"Mod4",		Mod4Mask,	0},
    {"M4",		Mod4Mask,	0},
    {"Mod5",		Mod5Mask,	0},
    {"M5",		Mod5Mask,	0},
    {"Double",		0,		DOUBLE},
    {"Triple",		0,		TRIPLE},
    {"Any",		0,		0},	/* Ignored: historical relic. */
    {NULL,		0,		0}
};
static Tcl_HashTable modTable;

/*
 * This module also keeps a hash table mapping from event names
 * to information about those events.  The structure, an array
 * to use to initialize the hash table, and the hash table are
 * all defined below.
 */

typedef struct {
    char *name;			/* Name of event. */
    int type;			/* Event type for X, such as
				 * ButtonPress. */
    int eventMask;		/* Mask bits (for XSelectInput)
				 * for this event type. */
} EventInfo;

/*
 * Note:  some of the masks below are an OR-ed combination of
 * several masks.  This is necessary because X doesn't report
 * up events unless you also ask for down events.  Also, X
 * doesn't report button state in motion events unless you've
 * asked about button events.
 */
#define MouseWheelEvent     (LASTEvent + 3)
#define MouseWheelMask         (1L << 28)


static EventInfo eventArray[] = {
    {"Key",		KeyPress,		KeyPressMask},
    {"KeyPress",	KeyPress,		KeyPressMask},
    {"KeyRelease",	KeyRelease,		KeyPressMask|KeyReleaseMask},
    {"Button",		ButtonPress,		ButtonPressMask},
    {"ButtonPress",	ButtonPress,		ButtonPressMask},
    {"ButtonRelease",	ButtonRelease,
	    ButtonPressMask|ButtonReleaseMask},
    {"Motion",		MotionNotify,
	    ButtonPressMask|PointerMotionMask},
    {"MouseWheel",	MouseWheelEvent,	MouseWheelMask},
    {(char *) NULL,	0,			0}
};


/*
 * The defines and table below are used to classify events into
 * various groups.  The reason for this is that logically identical
 * fields (e.g. "state") appear at different places in different
 * types of events.  The classification masks can be used to figure
 * out quickly where to extract information from events.
 */
#define KEY                     0x1
#define BUTTON                  0x2
#define MOTION                  0x4
#define CROSSING                0x8
#define FOCUS                   0x10
#define EXPOSE                  0x20
#define PROP                    0x8000
#define VIRTUAL                 0x20000
#define KEY_BUTTON_MOTION_VIRTUAL       (KEY|BUTTON|MOTION|VIRTUAL)



static int flagArray[TK_LASTEVENT] = {
   /* Not used */               0,
   /* Not used */               0,
   /* KeyPress */               KEY,
   /* KeyRelease */             KEY,
   /* ButtonPress */            BUTTON,
   /* ButtonRelease */          BUTTON,
   /* MotionNotify */           MOTION,
};



/*
 * Global variables used by the main program:
 */

static Tcl_Interp *interp;      /* Interpreter for this application. */
static Tcl_DString command;     /* Used to assemble lines of terminal input
                                 * into Tcl commands. */
static Tcl_DString line;        /* Used to read the next line from the
                                 * terminal input. */
static int tty;                 /* Non-zero means standard input is a
                                 * terminal-like device.  Zero means it's
                                 * a file. */

/*
 * Forward declarations for procedures defined later in this file.
 */

static void             Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));
static void             StdinProc _ANSI_ARGS_((ClientData clientData,
                            int mask));


#if defined(__WIN32__) || defined(_WIN32)
/* #   include "tkWinDefault.h"  */
#define NORMAL_BG       "SystemButtonFace"
#else
#   if defined(MAC_TCL)
/*#       include "tkMacDefault.h"  */
#define NORMAL_BG       "systemWindowBody"
#   else
/*       include "tkUnixDefault.h" */
#define NORMAL_BG       "#d9d9d9"
#   endif
#endif

#define BLACK           "Black"
#define WHITE           "White"

#define DEF_CANVAS_BG_COLOR             NORMAL_BG
#define DEF_CANVAS_BG_MONO              WHITE
#define DEF_CANVAS_BORDER_WIDTH         "0"
#define DEF_CANVAS_CONFINE              "1"
#define DEF_CANVAS_HEIGHT               "1c"
#define DEF_CANVAS_HIGHLIGHT_BG         NORMAL_BG
#define DEF_CANVAS_HIGHLIGHT            BLACK
#define DEF_CANVAS_HIGHLIGHT_WIDTH      "1"
#define DEF_CANVAS_RELIEF               "flat"
#define DEF_CANVAS_SCROLL_REGION        ""
#define DEF_CANVAS_TAKE_FOCUS           (char *) NULL
#define DEF_CANVAS_WIDTH                "1c"
#define DEF_CANVAS_X_SCROLL_CMD         ""
#define DEF_CANVAS_X_SCROLL_INCREMENT   "0"
#define DEF_CANVAS_Y_SCROLL_CMD         ""
#define DEF_CANVAS_Y_SCROLL_INCREMENT   "0"

#endif /* _XTALCANVAS */
